from math import pow
from math import log 

def compute_multiplet_rates_asymp(cell_num, sample_num, drop_num):
    no_drop_rate = (1 - 1 / drop_num)
    cells_per_sample = round(cell_num / sample_num)
    drop_with_cells = (1 - pow(no_drop_rate, cell_num)) * drop_num
    single_sample_drops = sample_num * (1 - pow(no_drop_rate, cells_per_sample) ) * drop_num * pow(no_drop_rate, (sample_num - 1) * cells_per_sample)
    singlet_drops = cell_num * pow(no_drop_rate, cell_num - 1)

    singlet_rate = singlet_drops / drop_with_cells
    single_sample_rate = single_sample_drops / drop_with_cells
    MSM_rate = 1 - single_sample_rate
    SSM_rate = single_sample_rate - singlet_rate
    return MSM_rate, SSM_rate, singlet_rate, drop_with_cells


def compute_relative_SSM_rate_asymp(cell_num, drop_num):
    no_drop_rate = (1 - 1 / drop_num)
    drop_with_cells = (1 - pow(no_drop_rate, cell_num)) * drop_num
    singlet_drops = cell_num * pow(no_drop_rate, cell_num - 1)

    return 1 - singlet_drops / drop_with_cells


def compute_relative_SSM_rate(SSM_rate, singlet_rate):
    return SSM_rate / singlet_rate


def get_min_hto_num(cell_num, drop_num, SSM_threshold, sample_num = 1):
    sample_num = 1
    while True:
        #(MSM_rate, SSM_rate, singlet_rate) = compute_multiplet_rates(cell_num, sample_num, drop_num)
        (MSM_rate, SSM_rate, singlet_rate, drop_with_cells) = compute_multiplet_rates_asymp(cell_num, sample_num, drop_num)
        relative_SSM_rate = compute_relative_SSM_rate(SSM_rate, singlet_rate)
        #print(relative_SSM_rate)
        if (relative_SSM_rate <= SSM_threshold):
            break;
        else:
            sample_num += 1

    return sample_num


def cell_num_estimator(a_num, captured_drop_num, capture_rate):
    estimated_drop_num = captured_drop_num / capture_rate
    base = 1 - 1 / estimated_drop_num
    power = 1 - a_num / captured_drop_num
    #print("power:", power)
    #print("base:", base)
    cell_num = log(power, base)
    return cell_num


def drop_num_estimator(a_num, b_num, shared_num):
    drop_num = a_num * b_num / shared_num 
    return drop_num 


def compute_shared_num(drop_num, A_num, B_num):
    A_rate = compute_mix_rate(drop_num, B_num) 
    #print("A_rate: ", A_rate)
    shared_num = A_rate *  A_num
    return shared_num


# Computes the rate of drops that have certain cells in them.
def compute_mix_rate(drop_num, cell_num):
    no_drop_rate = (1 - 1 / drop_num)
    cell_in_rate = 1 - pow(no_drop_rate, cell_num)
    return cell_in_rate


def compute_SSM_rate_with_cell_num(cell_num, drop_num):
    no_drop_rate = (1 - 1 / drop_num)
    singlet_drops = cell_num * pow(no_drop_rate, cell_num - 1)
    drop_with_cells = (1 - pow(no_drop_rate, cell_num)) * drop_num
    SSM_rate = 1 - singlet_drops / drop_with_cells
    return SSM_rate




# Just to suppress sklearn import imp warning
def warn(*args, **kwargs):
    pass
import warnings
warnings.warn = warn

import pandas as pd
from GMM_Demux import compute_venn
from GMM_Demux import estimator
from GMM_Demux import classify_drops
from GMM_Demux import GMM_IO
from sys import argv
import sys
import argparse
from tabulate import tabulate

def main():
    ####### Parsing parameters and preparing data #######
    parser = argparse.ArgumentParser()
       
    parser.add_argument('input_path', help = "The input path of mtx files from cellRanger pipeline.")
    parser.add_argument('hto_array', help = "Names of the HTO tags, separated by ','.")
    parser.add_argument('cell_num', help = "Estimated total number of cells across all HTO samples.")

    parser.add_argument("-f", "--full", help="Generate the full classification report. Require a path argument.", type=str)
    parser.add_argument("-s", "--simplified", help="Generate the simplified classification report. Require a path argument.", type=str)
    parser.add_argument("-o", "--output", help="The path for storing the Same-Sample-Droplets (SSDs). SSDs are stored in mtx format. Require a path argument.", type=str)
    parser.add_argument("-r", "--report", help="Store the data summary report. Require a file argument.", type=str)

    args = parser.parse_args()

    input_path = args.input_path
    hto_array = args.hto_array.split(',')
    estimated_total_cell_num = int(args.cell_num)
    if args.output:
        output_path = args.output
    else:
        output_path = "GMM_Demux_mtx"


    full_df, GMM_df = GMM_IO.read_cellranger(input_path, hto_array)
    confidence_threshold = 0.8 

    ####### Run classifier #######
    GEM_num = GMM_df.shape[0]
    sample_num = GMM_df.shape[1]

    base_bv_array = compute_venn.obtain_base_bv_array(sample_num)
    #print([int(i) for i in base_bv_array])
    (high_array, low_array) = classify_drops.obtain_arrays(GMM_df)

    # Obtain classification result
    GMM_full_df, class_name_ary = \
            classify_drops.classify_drops(base_bv_array, high_array, low_array, sample_num, GEM_num, GMM_df.index, GMM_df.columns.values)

    # Store classification results
    if args.full:
        print("Full classification result is stored in", args.full)
        classify_drops.store_full_classify_result(GMM_full_df, class_name_ary, args.full)

    if args.simplified:
        print("Simplified classification result is stored in", args.simplified)
        classify_drops.store_simplified_classify_result(GMM_full_df, class_name_ary, args.simplified, sample_num, confidence_threshold)

    ####### Estimate SSM #######
    # Count bad drops
    negative_num, unclear_num = classify_drops.count_bad_droplets(GMM_full_df, confidence_threshold)
    # Clean up bad drops
    purified_df = classify_drops.purify_droplets(GMM_full_df, confidence_threshold)

    # Select target samples
    SSD_idx = classify_drops.obtain_SSD_list(purified_df, sample_num)

    # Store SSD result
    print("***MSM-free droplets are stored in folder", output_path, "\n")
    SSD_df = GMM_IO.store_cellranger(full_df, SSD_idx, output_path)

    # Infer parameters
    (cell_num_ary, drop_num, capture_rate) = compute_venn.obtain_HTO_cell_n_drop_num(purified_df, base_bv_array, sample_num, estimated_total_cell_num, confidence_threshold)

    SSM_rate_ary = [estimator.compute_SSM_rate_with_cell_num(cell_num_ary[i], drop_num) for i in range(sample_num)]
    rounded_cell_num_ary = [round(cell_num) for cell_num in cell_num_ary]
    SSD_count_ary = classify_drops.get_SSD_count_ary(purified_df, SSD_idx, sample_num)
    count_ary = classify_drops.count_by_class(purified_df, base_bv_array)
    MSM_rate, SSM_rate, singlet_rate = compute_venn.gather_multiplet_rates(count_ary, SSM_rate_ary, sample_num)

    # Generate report
    full_report_dict = {
            "#Drops": round(drop_num),
            "Capture rate": "%5.2f" % (capture_rate * 100),
            "#Cells": sum(rounded_cell_num_ary),
            "Singlet": "%5.2f" % (singlet_rate * 100),
            "MSM": "%5.2f" % (MSM_rate * 100),
            "SSM": "%5.2f" % (SSM_rate * 100),
            "RSSM": "%5.2f" % (estimator.compute_relative_SSM_rate(SSM_rate, singlet_rate) * 100),
            "Negative": "%5.2f" % (negative_num / GMM_full_df.shape[0] * 100),
            "Unclear": "%5.2f" % (unclear_num / GMM_full_df.shape[0] * 100)
            }
    full_report_columns = [
            "#Drops",
            "Capture rate",
            "#Cells",
            "Singlet",
            "MSM",
            "SSM",
            "RSSM",
            "Negative",
            "Unclear"]

    full_report_df = pd.DataFrame(full_report_dict, index = ["Total"], columns=full_report_columns)



    print("==============================Full Report==============================")
    print(tabulate(full_report_df, headers='keys', tablefmt='psql'))
    print ("\n\n")
    print("==============================Per Sample Report==============================")
    sample_df = pd.DataFrame(data=[
        ["%d" % num for num in rounded_cell_num_ary],
        ["%d" % num for num in SSD_count_ary],
        ["%5.2f" % (num * 100) for num in SSM_rate_ary]
        ],
        columns = GMM_df.columns, index = ["#Cells", "#SSDs", "RSSM"])
    print(tabulate(sample_df, headers='keys', tablefmt='psql'))

    if args.report:
        print("\n\n***Summary report is stored in folder", args.report)
        with open(args.report, "w") as report_file:
            report_file.write("==============================Full Report==============================\n")
        with open(args.report, "a") as report_file:
            report_file.write(tabulate(full_report_df, headers='keys', tablefmt='psql'))
        with open(args.report, "a") as report_file:
            report_file.write("\n\n")
            report_file.write("==============================Per Sample Report==============================\n")
        with open(args.report, "a") as report_file:
            report_file.write(tabulate(sample_df, headers='keys', tablefmt='psql'))


if __name__ == "__main__":
    main()



from GMM_Demux import estimator
from GMM_Demux import classify_drops
import BitVector
import numpy as np

from sys import argv
import pandas as pd
from statistics import mean

# Returns the binary array representing all combinations of cells
def obtain_base_bv_array(sample_num):
    base_bv_array = []
    bv_array_segments = [0]

    for i in range(sample_num):
        base_bv_array.append(BitVector.BitVector(size = sample_num) )
        base_bv_array[-1][i] = 1

    bv_array_segments.append(len(base_bv_array))

    for i in range(1, sample_num):
        for j in range(bv_array_segments[-2], bv_array_segments[-1]):
            first_zero = sample_num - base_bv_array[j].reverse().next_set_bit()

            for k in range(first_zero, sample_num):
                base_bv_array.append(base_bv_array[j] | base_bv_array[k])
        bv_array_segments.append(len(base_bv_array))

    return [BitVector.BitVector(intVal = 0, size = sample_num)] + base_bv_array


def get_empty_bv(sample_num):
    return BitVector.BitVector(size = sample_num)


# Return true if the bit_pos th bit of bv is 1
def check_set_bit(bv, bit_pos, sample_num):
    mask = BitVector.BitVector(size = sample_num)
    mask[bit_pos] = 1
    
    return int(mask & bv) != 0


def gather_multiplet_rates(venn_values, SSM_rate_ary, sample_num):
    total_drops = 0
    total_singlets = 0
    total_MSMs = 0
    total_SSMs = 0

    for i in range(len(venn_values)):
        if i < sample_num:
            singlet_num = venn_values[i] * (1 - SSM_rate_ary[i])
            SSM_num = venn_values[i] - singlet_num

            total_SSMs += SSM_num
            total_singlets += singlet_num

        else:
            total_MSMs += venn_values[i]

        total_drops += venn_values[i]

    singlet_rate = total_singlets / total_drops
    SSM_rate = total_SSMs / total_drops
    MSM_rate = total_MSMs / total_drops

    return MSM_rate, SSM_rate, singlet_rate


def obtain_HTO_cell_n_drop_num(data_df, base_bv_array, sample_num, estimated_total_cell_num, confidence_threshold):
    hto_num_ary = []
    HTO_num_ary = []
    estimated_drop_num_ary = []

    # Obtain hto numbers
    for i in range(sample_num):
        hto_num = 0

        for j in range(len(base_bv_array)):
            if check_set_bit(base_bv_array[j], i, sample_num):
                hto_num += (data_df["Cluster_id"] == j).sum()

        hto_num_ary.append(hto_num)

    for hto_a_idx in range(sample_num):
        for hto_b_idx in range(hto_a_idx + 1, sample_num):

            shared_num = 0

            for j in range(len(base_bv_array)):
                if check_set_bit(base_bv_array[j], hto_a_idx, sample_num) and check_set_bit(base_bv_array[j], hto_b_idx, sample_num):
                    shared_num += (data_df["Cluster_id"] == j).sum()

            if shared_num > 0:
                estimated_drop_num = estimator.drop_num_estimator(hto_num_ary[hto_a_idx], hto_num_ary[hto_b_idx], shared_num)
                estimated_drop_num_ary.append(estimated_drop_num)


    captured_drop_num = mean(estimated_drop_num_ary)
    #print("Drop num ary:", estimated_drop_num_ary)
    #print("Total drop num:", drop_num)

    diff_num = 10000000000

    #for capture_rate in np.arange(1.0, 0.0, -0.005):
        #print("test: ", capture_rate)

    for capture_rate in np.arange(1.0, 0.0, -0.005):

        HTO_num_ary = []

        for hto_idx in range(sample_num):
            hto_num = hto_num_ary[hto_idx]
            cell_num = estimator.cell_num_estimator(hto_num, captured_drop_num, capture_rate)
            HTO_num_ary.append(cell_num)

        total_cell_num = sum(HTO_num_ary)
        #print("total_cell_num: ", total_cell_num)
        #print("abs(total_cell_num - estimated_total_cell_num): ", abs(total_cell_num - estimated_total_cell_num))

        if (abs(total_cell_num - estimated_total_cell_num) > diff_num):
            break

        diff_num = abs(total_cell_num - estimated_total_cell_num)


    #print(HTO_num_ary)
    return HTO_num_ary, captured_drop_num / capture_rate, capture_rate




from scipy.io import mmread, mmwrite
import pandas as pd
import os
from sys import argv
import gzip
from scipy import stats
import numpy as np
import os
from scipy.sparse import csr_matrix

pd.options.mode.chained_assignment = None
                                                                                                                        
def clr_norm(data_df):
    for hto in data_df.columns.values:
        compensated_values = data_df.loc[:,hto].values + 1
        gmean = stats.gmean(compensated_values)
        #print(gmean)

        data_df.loc[:,hto] = np.log(np.true_divide(compensated_values, gmean))

    return data_df


def read_cellranger(path, hto_array):
    mtx_file = gzip.open(os.path.join(path, 'matrix.mtx.gz'), 'r')
    cell_matrix = (mmread(mtx_file))
    cell_matrix = cell_matrix.todense()

    name_file = gzip.open(os.path.join(path, 'barcodes.tsv.gz'), 'rt')
    cell_names = name_file.read().splitlines()
    cell_names = [name for name in cell_names]

    feature_file = gzip.open(os.path.join(path, 'features.tsv.gz'), 'rt')
    features = feature_file.read().splitlines()
    simple_features = [(feature.split('\t')[1] if (len(feature.split('\t')) > 1) else feature) for feature in features]

    full_df = pd.DataFrame(cell_matrix, simple_features, cell_names).T
    #print(full_df)

    data_df = full_df[hto_array]

    data_df = clr_norm(data_df)

    full_df.columns = features

    return full_df, data_df


def store_cellranger(data_df, SSD_idx, path):
    if not os.path.exists(path):
        os.makedirs(path)

    mtx_file = gzip.open(os.path.join(path, 'matrix.mtx.gz'), 'w')

    SSD_df = data_df.loc[SSD_idx,:]
    mmwrite(mtx_file, csr_matrix(SSD_df.T.values) )

    #print(SSD_df)

    feature_file = gzip.open(os.path.join(path, 'features.tsv.gz'), 'wt')
    for feature in SSD_df.columns.values:
        feature_file.write(feature + "\n")

    cell_file = gzip.open(os.path.join(path, 'barcodes.tsv.gz'), 'wt')
    for name in SSD_df.index.values:
        cell_file.write(name + "\n")



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="GMM_Demux",
    version="0.0.3",
    author="Hongyi Xin",
    author_email="gohongyi@gmail.com",
    description="A multiplet removal tool for processing cell hashing data",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/CHPGenetics/GMM-demux",
    packages=setuptools.find_packages(),
    entry_points={
        'console_scripts': [
            'GMM-demux = GMM_Demux.GMM_Demux:main'
        ]
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3',
    install_requires=[
        "pandas",
        "numpy",
        "scipy",
        "tabulate",
        "argparse",
        "statistics",
        "BitVector",
        "sklearn"
    ]
)



import numpy as np

import pandas as pd
from scipy import stats
from sys import argv
from sklearn.mixture import GaussianMixture
import os

from GMM_Demux import check_multi_comp
from GMM_Demux import compute_venn


def obtain_arrays(data):
    gmm = []
    high_array = []
    low_array = []

    for i in range(data.shape[1]):
        X = data.iloc[:,i].values[:, np.newaxis]

        # GMM values
        gmm.append(GaussianMixture(2).fit(X))
        x = np.linspace(-6, 6, 1000)[:, np.newaxis]
        logprob= gmm[-1].score_samples(x)
        responsibilities = gmm[-1].predict_proba(x)
        pdf = np.exp(logprob)
        pdf_individual = responsibilities * pdf[:, np.newaxis]
        #print(pdf_individual)

        #print(gmm[-1].means_)

        # Extract prob
        high_idx = np.argmax(gmm[-1].means_, axis=0)[0]
        post_prob = gmm[-1].predict_proba(X)
        high_array.append(post_prob[np.arange(post_prob.shape[0]), np.full(post_prob.shape[0], high_idx)])
        low_array.append(np.full(post_prob.shape[0], 1.0) - high_array[-1])

    return high_array, low_array


def classify_drops(base_bv_array, high_array, low_array, sample_num, GEM_num, index, column):

    classified_ary = np.full(GEM_num, 0)
    confidence_ary = np.full(GEM_num, 0.0)

    class_name_array = ["negative"]
    all_idx_ary = []

    for i in range(sample_num):
        all_idx_ary.append(i)

    # Detailed classification
    for i in range(len(base_bv_array)):
        bv = base_bv_array[i]
        high_idx_ary = []
        name = ""

        for j in range(sample_num):
            if compute_venn.check_set_bit(bv, j, sample_num):
                high_idx_ary.append(j)
                name += (column[j] + "-")
        
        if name != "":
            name = name[:-1]
            class_name_array.append(name)

        tmp_confidence_ary = check_multi_comp.compute_confidence(high_array, low_array, high_idx_ary, all_idx_ary)
        update_idx = (tmp_confidence_ary > confidence_ary).nonzero()[0]
        confidence_ary[update_idx] = tmp_confidence_ary[update_idx]
        classified_ary[update_idx] = i


    GMM_array = np.column_stack(tuple([classified_ary, confidence_ary]))
    GMM_full_df = pd.DataFrame(data=GMM_array, index = index, columns = ["Cluster_id", "Confidence"])
    GMM_full_df["Cluster_id"] = GMM_full_df["Cluster_id"].astype(int)

    return GMM_full_df, class_name_array


# Store full classification result
def store_full_classify_result(data_df, class_name_array, path):

    if not os.path.exists(path):
        os.makedirs(path)

    classify_file_name = os.path.join(path, "GMM_full.csv")
    config_file_name = os.path.join(path, "GMM_full.config")

    data_df.to_csv(classify_file_name)

    with open(config_file_name, 'w') as f:
        for i in range(len(class_name_array)):
            f.write("%s, %s\n" % (i, class_name_array[i]))



# Store simplified classification result
def store_simplified_classify_result(data_df, class_name_array, path, sample_num, confidence_threshold):
    simplified_df = data_df.copy()

    if not os.path.exists(path):
        os.makedirs(path)

    #print(simplified_df)

    MSM_idx = data_df.index[(data_df["Cluster_id"] > sample_num).nonzero()[0]]
    #print(MSM_idx)
    simplified_df.loc[MSM_idx, "Cluster_id"] = sample_num + 1
    unclear_idx = data_df.index[(data_df["Confidence"] < confidence_threshold).nonzero()[0]]
    #print(unclear_idx)
    simplified_df.loc[unclear_idx, "Cluster_id"] = sample_num + 2

    #print(simplified_df)

    classify_file_name = os.path.join(path, "GMM_simplified.csv")
    config_file_name = os.path.join(path, "GMM_simplified.config")

    simplified_df.to_csv(classify_file_name)

    simplified_name_array = [class_name_array[i] for i in range(sample_num + 1)]
    simplified_name_array.append("MSM")
    simplified_name_array.append("Unclear")

    with open(config_file_name, 'w') as f:
        for i in range(len(simplified_name_array)):
            f.write("%s, %s\n" % (i, simplified_name_array[i]))


def purify_droplets(data_df, confidence_threshold):
    drop_idx = data_df.index[((data_df["Confidence"] < confidence_threshold) | (data_df["Cluster_id"] == 0)).nonzero()[0]]
    #print((data_df["Cluster_id"] == 0).nonzero()[0])
    purified_df = data_df.drop(drop_idx)
    return purified_df


def count_bad_droplets(data_df, confidence_threshold):
    negative_num = (data_df["Cluster_id"] == 0).sum()
    unclear_num = (data_df["Confidence"] < confidence_threshold).sum()
    return negative_num, unclear_num


def obtain_SSD_list(data_df, sample_num):
    SSD_idx = data_df.index[data_df["Cluster_id"] <= sample_num]
    return SSD_idx


def count_by_class(data_df, base_bv_array):
    count_ary = []
    for i in range(1, len(base_bv_array) + 1):
        count_ary.append((data_df["Cluster_id"] == i).sum())

    return count_ary


def get_SSD_count_ary(data_df, SSD_idx, sample_num):
    SSD_df = data_df.loc[SSD_idx,:]
    SSD_count_ary = []

    for i in range(1, sample_num + 1):
        SSD_count_ary.append((SSD_df["Cluster_id"] == i).sum())

    return SSD_count_ary


